#include <sys/ports.h>
#include <stdio.h>
#include <stdarg.h>
#include <stddef.h>
#include <sys/sio.h>
#include <stdlib.h>
#include <string.h>
#include <VTioroutines.h>

// Delay ms*100 microseconds
// This delay is not exact by any means
void VTdelay_100micros(unsigned short ms)
{
  unsigned char old_prescale_val;
  unsigned short tcnt, tend, tbegin;

	// save the old timer prescale value
	old_prescale_val = _io_ports[M6811_TMSK2];
	_io_ports[M6811_TMSK2] &= 0xFC;

	// Timer runs at 2 Mhz
	// One tick is 0.5 microseconds
	// we need 2000 ticks for 1 millisecond
	// or 200 ticks for 100 microseconds
  while (ms > 0)
    {
		tbegin = ((unsigned volatile short *) &(_io_ports[M6811_TCNT]))[0];
		tend = tbegin + 200;
		if (tend < tbegin) {
			// overflow
			while ((tcnt = ((unsigned volatile short *) &(_io_ports[M6811_TCNT]))[0]) > tbegin);
			while ((tcnt = ((unsigned volatile short *) &(_io_ports[M6811_TCNT]))[0]) < tend);
		} else {
			while ((tcnt = ((unsigned volatile short *) &(_io_ports[M6811_TCNT]))[0]) < tend);
		}
		while (((unsigned volatile short *) &(_io_ports[M6811_TCNT]))[0] < 2000);
      ms--;
    }
	// restore the old timer prescale value
	_io_ports[M6811_TMSK2] = old_prescale_val;
}

extern "C" void sbrk_init();
extern "C" void malloc_init();
void VT_init()
{
	sbrk_init();
	malloc_init();
	VTlcd_init();
}

// Turns LEDs on or off
// If Bit k is '1', then PBk is on
// If Bit k is '0', then PBk is off
void VTled(unsigned char leds)
{
	_io_ports[M6811_PORTB] = leds;
}

// Return the value of the switches 1-8 in a single byte
// If bit k of the byte is '1', then switch(k+1) is on
// If bit k of the byte is '0', then switch(k+1) is off
// NOTE: Bit 0 is also '1' if SW4 is pushed
// NOTE: Bit 1 is also '1' if SW3 is pushed
unsigned char VTswitches()
{
	return(~(_io_ports[M6811_PORTC]));
}

bool VTbuttonSW2()
{
	if (_io_ports[M6811_PORTA] & 0x01) {
		return(false);
	} else {
		return(true);
	}
}

unsigned int VTadc()
{
	_io_ports[M6811_PORTD] = 0xff;
	_io_ports[M6811_DDRD] = 0xfc;
	_io_ports[M6811_OPTION] = _io_ports[M6811_OPTION] | 0x80; // enable ADC
	_io_ports[M6811_ADCTL] = 7; // SINGLE_MODE + NO_SCAN + CH7
	while (!(_io_ports[M6811_ADCTL] & 0x80)) {
	}
	return((_io_ports[M6811_ADR1]+_io_ports[M6811_ADR2]+_io_ports[M6811_ADR3]+
		_io_ports[M6811_ADR4])/4);
}

unsigned int VTkeypad()
{
	unsigned short tval;
	unsigned int acc = 0;
	unsigned char mask = 0xfb;
	unsigned char curval = 0x04;
	int	i;
	unsigned int smask1 = 0x0001;
	unsigned int smask2 = 0x0010;
	unsigned int smask3 = 0x0100;
	unsigned int smask4 = 0x1000;

	_io_ports[M6811_DDRD] = 0xfe;
	for (i=0;i<4;i++) {
		tval = _io_ports[M6811_PORTD];
		tval = tval | 0x3c;
		tval = tval & mask;
		_io_ports[M6811_PORTD] = tval;
		VTdelay_100micros(1);
		tval = _io_ports[M6811_PORTE] & 0x0f;
		if (!(tval & 0x01)) {
			acc = acc | smask1;
		}
		if (!(tval & 0x02)) {
			acc = acc | smask2;
		}
		if (!(tval & 0x04)) {
			acc = acc | smask3;
		}
		if (!(tval & 0x08)) {
			acc = acc | smask4;
		}
		smask1 = smask1 << 1;
		smask2 = smask2 << 1;
		smask3 = smask3 << 1;
		smask4 = smask4 << 1;
		mask = mask | curval;
		curval = curval << 1;
		mask = mask & (~curval);
	}
	return(acc);
}

void VTbeep()
{
	int i;

	for (i=0;i<1000;i++) {
		_io_ports[M6811_PORTA] = _io_ports[M6811_PORTA] | 0x20;
		VTdelay_100micros(20);
		_io_ports[M6811_PORTA] = _io_ports[M6811_PORTA] & 0xDF;
	}
}

// LCD device names for bits in *LCD port
static const unsigned char  NEN = 0xFD;
static const unsigned char  EN = 0x02;
static const unsigned char  NRS = 0xFE;
static const unsigned char  RS = 0x01;
static const unsigned char  NOT_RESET = 0xF7;

static unsigned char lcdVal;

// private functions only visible in this file
// names taken from lcd1.asm on Fox11 CD buffalo directory
static void wrt_pulse(unsigned char data, bool rst = false);
static void enable_pulse(void);
static void outins(unsigned char *data, unsigned char len, bool rst = false);
static void sel_inst(void);
static void sel_data(void);
static void msg_out(char *str, unsigned char len);

void VTlcd_init(void)
{
    // initialization codes
    unsigned char inidsp1[] = {0x33, 0x32};
    unsigned char inidsp2[] = {0x28, 0x06, 0x0C, 0x01};

	VTdelay_100micros(200);
   lcdVal = 0;
	_io_ports[M6811_PORTF] = lcdVal;

    outins(inidsp1, 2, true);
    outins(inidsp2, 4, false);
}

void VTlcd_print(unsigned char line, unsigned char col, 
               char *str, unsigned char len)
{
    sel_inst();
    if (line == 0)
        wrt_pulse(0x80+col,false);
    else
        wrt_pulse(0xC0+col,false);
    sel_data();
    msg_out(str, len);
}

static void outins(unsigned char *data, unsigned char len, bool rst)
{
    sel_inst();

    for (unsigned char i = 0; i<len; i++)
    {
        wrt_pulse(data[i], rst);
			VTdelay_100micros(50);
    }
}

static void msg_out(char *str, unsigned char len)
{
    if (len)
        for (unsigned char i=0; i<len; i++)
            wrt_pulse(str[i],false);
    else
        while (*str)
            wrt_pulse(*str++,false);
}

static void wrt_pulse(unsigned char data, bool rst)
{
	unsigned char temp1, temp2;

	temp1 = data & 0xf0;
   lcdVal &= 0x0f;
   lcdVal |= temp1;
	temp2 = lcdVal & NOT_RESET;
	_io_ports[M6811_PORTF] = temp2;
   enable_pulse();
   if (rst) VTdelay_100micros(50);
	temp1 = data << 4;
   lcdVal &= 0x0f;
   lcdVal |= temp1;
	temp2 = lcdVal & NOT_RESET;
	_io_ports[M6811_PORTF] = temp2;
   enable_pulse();
}

static void enable_pulse(void)
{
	unsigned char temp1;

   lcdVal |= EN;
	temp1 = lcdVal & NOT_RESET;
	_io_ports[M6811_PORTF] = temp1;
   lcdVal &= NEN;
	temp1 = lcdVal & NOT_RESET;
	_io_ports[M6811_PORTF] = temp1;
	VTdelay_100micros(1);
}

static void sel_data(void)
{
	unsigned char temp1;

   lcdVal |= RS;
	temp1 = lcdVal & NOT_RESET;
	_io_ports[M6811_PORTF] = temp1;
}

static void sel_inst(void)
{
	unsigned char temp1;

   lcdVal &= NRS;
	temp1 = lcdVal & NOT_RESET;
	_io_ports[M6811_PORTF] = temp1;
}
